"""
Module defining login utility functions

Copyright (C) 2017 ERT Inc.
"""
__author__ = "Brandon J. Van Vaerenbergh <brandon.vanvaerenbergh@noaa.gov>"

import logging

from onelogin.saml2.auth import OneLogin_Saml2_Auth
from onelogin.saml2.utils import OneLogin_Saml2_Utils

from . import metadata

logger = logging.getLogger(__name__)

class LoginInvalid(Exception):
    # Single-Sign-On login attempt is invalid
    pass

class LoginSAMLResponseProcessingError(ValueError):
    # problem with response from remote Single-Sign-On authority
    pass

def login(host, port, url_path, settings_path, cert_path, key_path, get_data=dict(), post_data=dict()):
    """
    Return a dict representing attributes of a logged-in user

    Attributes for the logged in user are extracted from the valid
    SAML2 HTTP response data, sent by the remote Single-Sign-On server.

    Keyword Parameters:
    host  -- String, representing Hostname of server using remote SSO
    port  -- String, representing Port the remote SSO connects to
    url_path  -- String, representing the path component of the URL
      you provided to the remote SSO. E.g.: '/my_service/saml/' if SSO
      is configured to send login successes to
      'https://my.fake_domain/my_service/saml/?acs'
    settings_path  -- String, representing the path to your OneLogin
      settings.json file
    cert_path  -- String,representing path to the OpenSSL PEM formatted
      certificate file identifying the server which uses remote SSO
    key_path  -- String, representing path to the OpenSSL private key
      file used to sign cert_path certifiate file
    get_data  -- dict,representing SAML2 HTTP GET url parameters [E.g.:
      'SAMLResponse']. If post_data parameter is unspecified,a get_data
      parameter value *must* be provided. (Default: empty dict)
    post_data  -- dict, representing SAML2 HTTP POST parameters [E.g.:
      'SAMLResponse']. If get_data parameter is unspecified, a
      post_data parameter value must be provided. (Default: empty dict)

    Exceptions:
    LoginInvalid  -- raised if login failed
    LoginSAMLResponseProcessingError  -- raised if remote response was
      unusable
    """
    req =  { # prepare OneLogin SAML authentication request
        'http_host': host,
        'server_port': port,
        'script_name': url_path,
        'get_data': get_data.copy(),
        'post_data': post_data.copy()
    }
    with open(cert_path) as cert_file, open(key_path) as key_file:
        cert = cert_file.read()
        key = key_file.read()
    # process
    auth = OneLogin_Saml2_Auth(req, metadata.get_settings_json(settings_path, cert, key))
    auth.process_response()
    errors = auth.get_errors()
    if not errors:
        if auth.is_authenticated():
            # successful login!
            # perform requested SAML login action (if any)
            if ('RelayState' in req['post_data'] and
              OneLogin_Saml2_Utils.get_self_url(req) != req['post_data']['RelayState']):
                logger.info('Redirecting: '+req['post_data']['RelayState'])
                auth.redirect_to(req['post_data']['RelayState'])
            else:
                for attr_name in auth.get_attributes().keys():
                    logger.info('{} ==> {}'.format(attr_name, '|| '.join(auth.get_attributes()[attr_name])))
            return auth.get_attributes()
        else:
          raise LoginInvalid('Not authenticated')
    else:
        raise LoginSAMLResponseProcessingError("{}".format(', '.join(errors)))



"""
Module defining SAML2 Service Provider metadata utility functions

Copyright (C) 2017 ERT Inc.
"""
__author__ = "Brandon J. Van Vaerenbergh <brandon.vanvaerenbergh@noaa.gov>"

import json

from onelogin.saml2.auth import OneLogin_Saml2_Auth
from onelogin.saml2.utils import OneLogin_Saml2_Utils
from onelogin.saml2.errors import OneLogin_Saml2_Error

class EntityDescriptorSettingsError(OneLogin_Saml2_Error):
    # Raised on EntityDescriptor metadata generation failure
    pass

class EntityDescriptorValidationError(ValueError):
    # Raised on EntityDescriptor metadata XML validation failure
    pass

def get_entity_descriptor(settings_path, cert_path, key_path):
    """
    Return Service Provider metadata & content type,as tuple of Strings

    Keyword Parameters:
    settings_path  -- String, representing the path to your OneLogin
      settings.json file
    cert_path  -- String,representing path to the OpenSSL PEM formatted
      certificate file identifying the server which uses remote SSO
    key_path  -- String, representing path to the OpenSSL private key
      file used to sign cert_path certifiate file

    Exceptions:
    EntityDescriptorSettingsError  -- raised when contents of the
      settings_path file are invalid
    EntityDescriptorValidationError  -- raised if settings generate XML
      with an invalid SAML2 EntityDescriptor element

    >>> from pprint import pprint
    >>> import tempfile
    >>> test_f = tempfile.NamedTemporaryFile
    >>> with test_f() as settings, test_f() as cert, test_f() as key:
    ...     settings.write(b'''{
    ...     "strict": true,
    ...     "debug": true,
    ...     "sp": {
    ...         "entityId": "https://my_server.fake_domain/saml/metadata/",
    ...         "assertionConsumerService": {
    ...             "url": "https://my_server.gov/saml/?acs",
    ...             "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"
    ...         },
    ...         "singleLogoutService": {
    ...             "url": "https://my_server.gov/saml/?sls",
    ...             "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect"
    ...         },
    ...         "attributeConsumingService": {
    ...                 "serviceName": "Test Name",
    ...                 "serviceDescription": "A Test Service",
    ...                 "requestedAttributes": [
    ...                     {
    ...                         "name": "TestName",
    ...                         "isRequired": false,
    ...                         "nameFormat": "",
    ...                         "friendlyName": "",
    ...                         "attributeValue": []
    ...                     }
    ...                 ]
    ...         },
    ...         "NameIDFormat": "urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified",
    ...         "x509cert": "",
    ...         "privateKey": ""
    ...     },
    ...     "idp": {
    ...         "entityId": "noaa-online-idp",
    ...         "singleSignOnService": {
    ...             "url": "https://fake_sso_server.gov/sso/noaa-online-idp",
    ...             "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect"
    ...         },
    ...         "singleLogoutService": {
    ...             "url": "https://fake_sso_server.gov/slo/noaa-online-idp",
    ...             "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect"
    ...         },
    ...         "x509cert": "invalid_cert"
    ...     }
    ... }''') #write out 1554 bytes
    ...     settings.seek(0) #reset file object, to beginning of file
    ...     test_out = get_entity_descriptor(settings.name, cert.name, key.name)
    1554
    0
    >>> len(test_out) #validate tuple length
    2
    >>> test_out[1] #check last element
    'text/xml'
    >>> #now, check first element
    >>> pprint(test_out[0]) # doctest: +ELLIPSIS
    '<?xml version="1.0"?>\\n'
    '<md:EntityDescriptor xmlns:md="urn:oasis:names:tc:SAML:2.0:metadata"\\n'
    '                     validUntil="...Z"\\n'
    '                     cacheDuration="PT604800S"\\n'
    '                     '
    'entityID="https://my_server.fake_domain/saml/metadata/">\\n'
    '    <md:SPSSODescriptor AuthnRequestsSigned="false" '
    'WantAssertionsSigned="false" '
    'protocolSupportEnumeration="urn:oasis:names:tc:SAML:2.0:protocol">\\n'
    '        <md:SingleLogoutService '
    'Binding="urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect"\\n'
    '                                Location="https://my_server.gov/saml/?sls" '
    '/>\\n'
    '        '
    '<md:NameIDFormat>urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified</md:NameIDFormat>\\n'
    '        <md:AssertionConsumerService '
    'Binding="urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"\\n'
    '                                     '
    'Location="https://my_server.gov/saml/?acs"\\n'
    '                                     index="1" />\\n'
    '        <md:AttributeConsumingService index="1">\\n'
    '            <md:ServiceName xml:lang="en">Test Name</md:ServiceName>\\n'
    '            <md:ServiceDescription xml:lang="en">A Test '
    'Service</md:ServiceDescription>\\n'
    '            <md:RequestedAttribute Name="TestName" />\\n'
    '        </md:AttributeConsumingService>\\n'
    '    </md:SPSSODescriptor>\\n'
    '\\n'
    '\\n'
    '</md:EntityDescriptor>'
    """
    fake_request = dict()
    with open(cert_path) as cert_file, open(key_path) as key_file:
        cert = cert_file.read()
        key = key_file.read()
    try:
        auth = OneLogin_Saml2_Auth(fake_request, get_settings_json(settings_path, cert, key))    
        s = auth.get_settings()
        md = s.get_sp_metadata()
        errors = s.validate_metadata(md)
        if errors:
            raise EntityDescriptorValidationError(errors)
        content_type = 'text/xml'
        return (md, content_type)
    except OneLogin_Saml2_Error as e:
        raise EntityDescriptorSettingsError(e) from e

def get_settings_json(settings_path, cert_string, key_string):
    """
    Return String, representing OneLogin SAML settings.json

    Keyword Parameters:
    settings_path  -- String, representing the path to your OneLogin
      settings.json file
    cert_path  -- String,representing path to the OpenSSL PEM formatted
      certificate file identifying the server which uses remote SSO
    key_path  -- String, representing path to the OpenSSL private key
      file used to sign cert_path certifiate file

    >>> from pprint import pprint
    >>> import tempfile
    >>> with tempfile.NamedTemporaryFile() as settings:
    ...     settings.write(b'''{
    ...     "strict": true,
    ...     "debug": true,
    ...     "sp": {
    ...         "entityId": "https://my_server.fake_domain/saml/metadata/",
    ...         "assertionConsumerService": {
    ...             "url": "https://my_server.fake_domain/saml/?acs",
    ...             "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST"
    ...         },
    ...         "singleLogoutService": {
    ...             "url": "https://my_server.fake_domain/saml/?sls",
    ...             "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect"
    ...         },
    ...         "attributeConsumingService": {
    ...                 "serviceName": "Test Name",
    ...                 "serviceDescription": "A Test Service",
    ...                 "requestedAttributes": [
    ...                     {
    ...                         "name": "TestName",
    ...                         "isRequired": false,
    ...                         "nameFormat": "",
    ...                         "friendlyName": "",
    ...                         "attributeValue": []
    ...                     }
    ...                 ]
    ...         },
    ...         "NameIDFormat": "urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified",
    ...         "x509cert": "",
    ...         "privateKey": ""
    ...     },
    ...     "idp": {
    ...         "entityId": "noaa-online-idp",
    ...         "singleSignOnService": {
    ...             "url": "https://fake.test_domain/sso/noaa-online-idp",
    ...             "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect"
    ...         },
    ...         "singleLogoutService": {
    ...             "url": "https://fake.test_domain/slo/noaa-online-idp",
    ...             "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect"
    ...         },
    ...         "x509cert": "invalid_cert"
    ...     }
    ... }''') #write out 1564 bytes
    ...     settings.seek(0) #reset file object, to beginning of file
    ...     test_out = get_settings_json(settings.name, 'foo', 'bar')
    1564
    0
    >>> pprint(test_out)
    {'debug': True,
     'idp': {'entityId': 'noaa-online-idp',
             'singleLogoutService': {'binding': 'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect',
                                     'url': 'https://fake.test_domain/slo/noaa-online-idp'},
             'singleSignOnService': {'binding': 'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect',
                                     'url': 'https://fake.test_domain/sso/noaa-online-idp'},
             'x509cert': 'invalid_cert'},
     'sp': {'NameIDFormat': 'urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified',
            'assertionConsumerService': {'binding': 'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST',
                                         'url': 'https://my_server.fake_domain/saml/?acs'},
            'attributeConsumingService': {'requestedAttributes': [{'attributeValue': [],
                                                                   'friendlyName': '',
                                                                   'isRequired': False,
                                                                   'name': 'TestName',
                                                                   'nameFormat': ''}],
                                          'serviceDescription': 'A Test Service',
                                          'serviceName': 'Test Name'},
            'entityId': 'https://my_server.fake_domain/saml/metadata/',
            'privateKey': 'bar',
            'singleLogoutService': {'binding': 'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect',
                                    'url': 'https://my_server.fake_domain/saml/?sls'},
            'x509cert': 'foo'},
     'strict': True}
    """
    with open(settings_path) as settings_file:
        settings_string = settings_file.read()
    settings = json.loads(settings_string)
    sp = settings['sp']
    sp['x509cert'] = cert_string
    sp['privateKey'] = key_string
    return settings



"""
Authentication testcases

Copyright (C) 2017 ERT Inc.
"""
__author__ = "Brandon J. Van Vaerenbergh <brandon.vanvaerenbergh@noaa.gov>"

import doctest

from . import (
    metadata
    ,sso
)

def load_tests(loader, tests, ignore):
    """
    integrate Doctests with unittest test descovery API

    per: https://docs.python.org/3/library/doctest.html#unittest-api
    """
    tests.addTests(doctest.DocTestSuite(metadata))
    tests.addTests(doctest.DocTestSuite(sso))
    return tests



"""A simple setuptools based setup module

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject

Copyright (C) 2016-2017 The Python Packaging Authority, ERT Inc.
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path
import logging

here = path.abspath(path.dirname(__file__))

logger = logging.getLogger(__name__)

# Get the long description from the README file
try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst').replace("\r","")
except (OSError, ImportError):
    logger.warning("Pandoc not found. Long_description conversion failure.")
    # use raw Markdown text
    logger.warning("Falling back to raw README.md text, for long_description.")
    with open(path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()


# Versions should comply with PEP440.  For a discussion on single-sourcing
# the version across setup.py and the project code, see
# https://packaging.python.org/en/latest/single_source_version.html
__version__ = '0.1.0rc1'

setup(
    name='pyicam',
    version=__version__,
    description='SAML 2.0 client authentication library',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/nwfsc-data/PyICAM',

    # Author details
    author='NOAA Northwest science Fishery Resource Analysis and Montoring (FRAM) Data team',
    author_email='nwfsc.fram.data@noaa.gov',

    # Choose your license
    license='CC0-1.0',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Systems Administration :: Authentication/Directory',

        # Pick your license as you wish (should match "license" above)
        'License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    # What does your project relate to?
    keywords='saml saml2 sso icam noaa',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['python3-saml'],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': ['twine','pypandoc'],
        'test': ['coverage'],
    },
)

